#!BPY

"""
Name: 'Ultimate 3D model file format (*.u3d)'
Blender: 247
Group: 'Export'
Tip: 'Exports models (geometry, materials and animations) to the Ultimate 3D model file format (*.u3d).'
"""

__author__="Christoph Peters";
__url__="http://Ultimate3D.org";
__version__="1.05";
__bpydoc__="""Exports geometry, materials and armatures with animations and constraints of models in Blender to the Ultimate 3D model file format.""";

# ----------------------------- LICENSE -----------------------------
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Include the needed components of the Blender API
import Blender
from Blender import Scene
from Blender import Group
from Blender import Object
from Blender import Material
from Blender import Mathutils
from Blender import Mesh
from Blender import Armature
from Blender import Ipo
from Blender import Draw
from Blender import Window
# struct is needed to bring data into a binary format
import struct
# The math module is needed for the usual mathematical operations
import math
# This module is useful to copy files
import shutil
# This module is needed to create the directory to which the textures are output
import os

class C3DVector:
    """This class represents simple 3D vectors"""

    def __init__(self,x,y,z):
        """Constructs the vector from three components"""
        self.x=x;
        self.y=y;
        self.z=z;

    def SaveData(self,File):
        """Saves the data of this vector to the given file"""
        File.write(struct.pack("fff",self.x,self.y,self.z));
    
    def SaveAsCompressedNormal(self,File):
        """Saves a normalized copy of this vector as 16-byte
           longitude/latitude vector."""
        self.Normalize();
        Longitude=-math.asin(self.y);
        CosLatitude=0.9999*self.z/max(0.0001,math.cos(Longitude));
        Latitude=math.acos(CosLatitude);
        Latitude*=(1.0-2.0*(self.x<0.0));
        Longitude*=20.292255;
        Latitude*=40.58451;
        File.write(struct.pack("bb",int(Longitude),int(Latitude)));

    def Normalize(self):
        """Normalizes this vector"""
        Length=math.sqrt(self.x*self.x+self.y*self.y+self.z*self.z);
        if(Length>0.0):
            InvLength=1.0/Length;
            self.x*=InvLength;
            self.y*=InvLength;
            self.z*=InvLength;
    
    x=0.0;
    y=0.0;
    z=0.0;


class CMatrix4x4:
    """This class represents a 4x4 transformation matrix."""

    def __init__(self,_11,_12,_13,_14,_21,_22,_23,_24,_31,_32,_33,_34,_41,_42,_43,_44):
        """This constructor initializes this matrix with the given entries."""
        self._11=_11; self._12=_12; self._13=_13; self._14=_14;
        self._21=_21; self._22=_22; self._23=_23; self._24=_24;
        self._31=_31; self._32=_32; self._33=_33; self._34=_34;
        self._41=_41; self._42=_42; self._43=_43; self._44=_44;

    def CreateFromBlenderMatrix(self,BlenderMatrix):
        """Copies the given Blender matrix to this matrix."""
        if(BlenderMatrix.colSize<4 or BlenderMatrix.rowSize<4):
            print("Failed to copy a Blender matrix object to a 4x4 matrix. The Blender matrix has a size of "+str(BlenderMatrix.rowSize)+"x"+str(BlenderMatrix.colSize)+".");
        else:
            self._11=BlenderMatrix[0][0]; self._12=BlenderMatrix[0][1]; self._13=BlenderMatrix[0][2]; self._14=BlenderMatrix[0][3];
            self._21=BlenderMatrix[1][0]; self._22=BlenderMatrix[1][1]; self._23=BlenderMatrix[1][2]; self._24=BlenderMatrix[1][3];
            self._31=BlenderMatrix[2][0]; self._32=BlenderMatrix[2][1]; self._33=BlenderMatrix[2][2]; self._34=BlenderMatrix[2][3];
            self._41=BlenderMatrix[3][0]; self._42=BlenderMatrix[3][1]; self._43=BlenderMatrix[3][2]; self._44=BlenderMatrix[3][3];

    def ApplyTranslation(self,BlenderVector):
        self._41=BlenderVector[0];
        self._42=BlenderVector[1];
        self._43=BlenderVector[2];

    def SaveData(self,File):
        """This function saves this matrix to a given binary file."""
        File.write(struct.pack("16f",self._11,self._12,self._13,self._14,self._21,self._22,self._23,self._24,self._31,self._32,self._33,self._34,self._41,self._42,self._43,self._44));

    def TransformVector(self,Vector):
        """This function returns a transformed version of the given vector."""
        return C3DVector(Vector.x*self._11+Vector.y*self._21+Vector.z*self._31+self._41,Vector.x*self._12+Vector.y*self._22+Vector.z*self._32+self._42,Vector.x*self._13+Vector.y*self._23+Vector.z*self._33+self._43);

    def TransformMatrix(self,Matrix):
        """This function transforms the given matrix by this matrix. The
           resulting matrix will describe the transformation, that can be
           achieved by using first the given matrix and after that this
           matrix."""
        return CMatrix4x4(
            self._11*Matrix._11+self._21*Matrix._12+self._31*Matrix._13+self._41*Matrix._14,
            self._12*Matrix._11+self._22*Matrix._12+self._32*Matrix._13+self._42*Matrix._14,
            self._13*Matrix._11+self._23*Matrix._12+self._33*Matrix._13+self._43*Matrix._14,
            self._14*Matrix._11+self._24*Matrix._12+self._34*Matrix._13+self._44*Matrix._14,
            
            self._11*Matrix._21+self._21*Matrix._22+self._31*Matrix._23+self._41*Matrix._24,
            self._12*Matrix._21+self._22*Matrix._22+self._32*Matrix._23+self._42*Matrix._24,
            self._13*Matrix._21+self._23*Matrix._22+self._33*Matrix._23+self._43*Matrix._24,
            self._14*Matrix._21+self._24*Matrix._22+self._34*Matrix._23+self._44*Matrix._24,
            
            self._11*Matrix._31+self._21*Matrix._32+self._31*Matrix._33+self._41*Matrix._34,
            self._12*Matrix._31+self._22*Matrix._32+self._32*Matrix._33+self._42*Matrix._34,
            self._13*Matrix._31+self._23*Matrix._32+self._33*Matrix._33+self._43*Matrix._34,
            self._14*Matrix._31+self._24*Matrix._32+self._34*Matrix._33+self._44*Matrix._34,
            
            self._11*Matrix._41+self._21*Matrix._42+self._31*Matrix._43+self._41*Matrix._44,
            self._12*Matrix._41+self._22*Matrix._42+self._32*Matrix._43+self._42*Matrix._44,
            self._13*Matrix._41+self._23*Matrix._42+self._33*Matrix._43+self._43*Matrix._44,
            self._14*Matrix._41+self._24*Matrix._42+self._34*Matrix._43+self._44*Matrix._44,
            );

    def Compute3x3Determinant(self):
        """This function computes the determinant of the 3x3 sub-matrix of
           this matrix and returns it."""
        return self._11*(self._22*self._33-self._32*self._23)-self._12*(self._21*self._33-self._31*self._23)+self._13*(self._21*self._32-self._31*self._22);

    def RotateAndScaleVector(self,Vector):
        """This function returns a rotated and scaled version of the given
           vector. Translation is not applied."""
        return C3DVector(Vector.x*self._11+Vector.y*self._21+Vector.z*self._31,Vector.x*self._12+Vector.y*self._22+Vector.z*self._32,Vector.x*self._13+Vector.y*self._23+Vector.z*self._33);

    def Print(self):
        """This function outputs this matrix to the console."""
        print("_11: "+str(self._11)+" _12: "+str(self._12)+" _13: "+str(self._13)+" _14: "+str(self._14));
        print("_21: "+str(self._21)+" _22: "+str(self._22)+" _23: "+str(self._23)+" _24: "+str(self._24));
        print("_31: "+str(self._31)+" _32: "+str(self._32)+" _33: "+str(self._33)+" _34: "+str(self._34));
        print("_41: "+str(self._41)+" _42: "+str(self._42)+" _43: "+str(self._43)+" _44: "+str(self._44));

    _11=1.0; _12=0.0; _13=0.0; _14=0.0;
    _21=0.0; _22=1.0; _23=0.0; _24=0.0;
    _31=0.0; _32=0.0; _33=1.0; _34=0.0;
    _41=0.0; _42=0.0; _43=0.0; _44=1.0;


# This global matrix brings a vector from Blender space to Ultimate 3D space
# by switching the y and the z axis. If skinning is used it will be set to
# the identity matrix, because in this case the root bone does the
# transformation.
BlenderToU3DSpace=CMatrix4x4(1.0,0.0,0.0,0.0, 0.0,0.0,1.0,0.0, 0.0,1.0,0.0,0.0, 0.0,0.0,0.0,1.0);


class CQuaternion:
    """This class handles a quaternion describing a rotation."""

    def __init__(self,x,y,z,w):
        """This constructor initializes this quaternion with the given entries
           and normalizes it. If x, y and z are zero an identity rotation is
           created."""
        LengthSq=x*x+y*y+z*z+w*w;
        if(LengthSq>0.0):
            InvLength=1.0/math.sqrt(LengthSq);
            self.x=x*InvLength;
            self.y=y*InvLength;
            self.z=z*InvLength;
            self.w=w*InvLength;
        else:
            self.x=0.0;
            self.y=0.0;
            self.z=0.0;
            self.w=1.0;

    def SaveData(self,File):
        """This function saves the data of this quaternion to the given file."""
        File.write(struct.pack("4f",self.x,self.y,self.z,self.w));

    def AppendRotation(self,RHS):
        """This function multiplies this quaternion by the given quaternion.
           Transforming a vector with the resulting quaternion will be
           equivalent to transforming it first with this quaternion and after
           that with the given quaternion."""
        x=self.x;
        y=self.y;
        z=self.z;
        w=self.w;
        self.x=RHS.w*x+w*RHS.x+RHS.y*z-RHS.z*y;
        self.y=RHS.w*y+w*RHS.y+RHS.z*x-RHS.x*z;
	self.z=RHS.w*z+w*RHS.z+RHS.x*y-RHS.y*x;
        self.w=RHS.w*w-RHS.x*x-RHS.y*y-RHS.z*z;

    def RotateVector(self,VectorX,VectorY,VectorZ):
        """This function returns a version of the given vector, to which the
           rotation described by this vector has been applied."""
        return C3DVector(
            2.0*(VectorX*(0.5-(self.y*self.y+self.z*self.z))+VectorY*(self.x*self.y-self.z*self.w)+VectorZ*(self.x*self.z+self.y*self.w)),
            2.0*(VectorX*(self.x*self.y+self.z*self.w)+VectorY*(0.5-(self.x*self.x+self.z*self.z))+VectorZ*(self.y*self.z-self.x*self.w)),
            2.0*(VectorX*(self.x*self.z-self.y*self.w)+VectorY*(self.y*self.z+self.x*self.w)+VectorZ*(0.5-(self.x*self.x+self.y*self.y)))
	);
        
    # The member variables, describing an identity rotation
    x=0.0;
    y=0.0;
    z=0.0;
    w=1.0;


class CTriangleIndices:
    """This class holds three vertex indices, describing a triangle."""

    def __init__(self,A,B,C):
        """Constructs the triangle index set from three indices"""
        self.A=A;
        self.B=B;
        self.C=C;

    def SaveData(self,File):
        """Saves the data of this triangle index set to the given file"""
        File.write(struct.pack("HHH",self.A,self.B,self.C));
    
    A=0;
    B=0;
    C=0;


class CColor:
    """ This class represents an r, g, b, a color (four floats)."""

    def __init__(self,r,g,b,a):
        self.r=r;
        self.g=g;
        self.b=b;
        self.a=a;

    def SaveData(self,File):
        File.write(struct.pack("4f",self.r,self.g,self.b,self.a));
        
    r=1.0;
    g=1.0;
    b=1.0;
    a=1.0;


class CHeader:
    """This class collects header information for the *.u3d file"""

    def __init__(self,RelevantObjects,MaterialList,nBone,nFrame):
        """This constructor collects header data from the given scene."""
        # Count the number of meshes
        for i in RelevantObjects:
            if(i.getType()=="Mesh"):
                self.nMesh+=1;
        if(nBone):
            self.nMesh=1;

        # Count the number of frames
        self.nFrame=max(1,nFrame);
        # Get the number of materials
        self.nMaterial=len(MaterialList);
        if(self.nMaterial==0):
            self.nMaterial=1;
        # Get the number of bones
        self.nBone=nBone;
        print("The scene contains "+str(nBone)+" bones.");


    def SaveData(self,File):
        """This function writes collected header data to the given file object"""
        # Write the null-terminated file format identification string
        File.write("U3DModelFile");
        File.write(struct.pack("b",0));
        # Write the file version
        File.write(struct.pack("f",1.0));
        #Write the mesh count (twice)
        File.write(struct.pack("HH",self.nMesh,self.nMesh));
        # Write the frame count
        File.write(struct.pack("L",self.nFrame));
        # Write the LoD count
        File.write(struct.pack("H",1));
        # Write the material count
        File.write(struct.pack("H",self.nMaterial));
        # Write the bone count
        File.write(struct.pack("H",self.nBone));
        # Write the cel-shading line strength and the cel-shading color
        File.write(struct.pack("5f",0.0,0.0,0.0,0.0,0.0));
        # Write whether this model uses alpha blending
        File.write(struct.pack("b",1));
        # Write whether this model uses vertex tweening
        File.write(struct.pack("b",0));

    # These variables save the collected header information
    nMesh=0;
    nFrame=0;
    nMaterial=0;
    nBone=0;


class SVertex:
    """Just a little helper struct to manage vertex data with texture
       coordinates."""

    def __init__(self,VertexPosition,VertexNormal,VertexTextureCoordinate):
        """This constructor initializes this vertex with the given data."""
        self.Position=VertexPosition;
        self.Normal=VertexNormal;
        self.TextureCoordinate=VertexTextureCoordinate;


class CMesh:
    """This class is used to collect and export mesh data"""

    def __init__(self,MeshObject,MaterialList,BoneList,MeshTransformation,RemoveDoubleVertices):
        """This constructor collects mesh data from the given mesh."""
        self.CleanUp();
        # Get the mesh name
        self.Name=MeshObject.name;
        # Get the vertex count and the vertices
        self.bTextureCoordinatesAvailable=MeshObject.vertexUV or MeshObject.faceUV;
        if(len(BoneList)==0):
            MeshToU3DSpace=BlenderToU3DSpace.TransformMatrix(MeshTransformation);
        else:
            MeshToU3DSpace=MeshTransformation;
        # Determine whether the mesh is partially solid (in this case the
        # creation of the vertex list is more complicated)
        Solid=False;
        for Face in MeshObject.faces:
            if(not Face.smooth):
                Solid=True;
                break;
        # If the texture coordinates and the normals are saved per vertex the
        # process is simple
        VerticesResorted=((MeshObject.vertexUV==False and MeshObject.faceUV) or Solid);
        if(not VerticesResorted):
            for v in MeshObject.verts:
                self.pVertexPosition.append(MeshToU3DSpace.TransformVector(C3DVector(v.co.x,v.co.y,v.co.z)));
                self.pVertexNormal.append(MeshToU3DSpace.RotateAndScaleVector(C3DVector(v.no.x,v.no.y,v.no.z)));
                if MeshObject.vertexUV:
                    self.pVertexTexCoord.append([v.uvco.x,1.0-v.uvco.y]);
                self.nVertex+=1;
        # If the texture coordinates or the normals are saved on per triangle
        # base get them from there
        else:
            # Warn the user, that removing vertices will take a while.
            if(RemoveDoubleVertices):
                print("Removing double vertices for the mesh \""+self.Name+"\". This process may take a while.");
            # This case is a little complicated. One vertex gets created for
            # every face vertex. Then the list gets reduced to unique
            # combinations of positions, texture coordinates and normals.
            # During this process a lookup table for creating the index buffer
            # gets created.
            pVertex=[];
            ppiTriangleVertex=[];
            if(RemoveDoubleVertices):
                nExpectedVertices=0;
                for i in range(len(MeshObject.faces)):
                    ppiTriangleVertex.append([]);
                    for j in range(len(MeshObject.faces[i].verts)):
                        nExpectedVertices+=1;
                        if(MeshObject.faces[i].smooth):
                            VertexNormal=MeshObject.faces[i].verts[j].no;
                        else:
                            VertexNormal=MeshObject.faces[i].no;
                        if(MeshObject.vertexUV):
                            Vertex=SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,MeshObject.faces[i].verts[j].uvco);
                        elif(MeshObject.faceUV):
                            Vertex=SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,MeshObject.faces[i].uv[j]);
                        else:
                            Vertex=SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,Mathutils.Vector([0.0,0.0]));
                        iUniqueVertex=0;
                        for UniqueVertex in pVertex:
                            if(Vertex.Position.x==UniqueVertex.Position.x and Vertex.Position.y==UniqueVertex.Position.y and Vertex.Position.z==UniqueVertex.Position.z and Vertex.Normal.x==UniqueVertex.Normal.x and Vertex.Normal.y==UniqueVertex.Normal.y and Vertex.Normal.z==UniqueVertex.Normal.z and Vertex.TextureCoordinate.x==UniqueVertex.TextureCoordinate.x and Vertex.TextureCoordinate.y==UniqueVertex.TextureCoordinate.y):
                                ppiTriangleVertex[i].append(iUniqueVertex);
                                break;
                            iUniqueVertex+=1;
                        else:
                            ppiTriangleVertex[i].append(len(pVertex));
                            pVertex.append(Vertex);
                            self.nVertex+=1;
                print("Removed "+str(nExpectedVertices-self.nVertex)+" vertices ("+str(100-int((100.0*self.nVertex)/nExpectedVertices))+"%).");
            else:
                for i in range(len(MeshObject.faces)):
                    ppiTriangleVertex.append([]);
                    for j in range(len(MeshObject.faces[i].verts)):
                        ppiTriangleVertex[i].append(len(pVertex));
                        if(MeshObject.faces[i].smooth):
                            VertexNormal=MeshObject.faces[i].verts[j].no;
                        else:
                            VertexNormal=MeshObject.faces[i].no;
                        if(MeshObject.faceUV):
                            pVertex.append(SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,MeshObject.faces[i].uv[j]));
                        elif(MeshObject.vertexUV):
                            pVertex.append(SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,MeshObject.faces[i].verts[j].uvco));
                        else:
                            pVertex.append(SVertex(MeshObject.faces[i].verts[j].co,VertexNormal,Mathutils.Vector([0.0,0.0])));
                        self.nVertex+=1;
            # Create the vertex data arrays
            for Vertex in pVertex:
                self.pVertexPosition.append(MeshToU3DSpace.TransformVector(C3DVector(Vertex.Position.x,Vertex.Position.y,Vertex.Position.z)));
                self.pVertexNormal.append(MeshToU3DSpace.RotateAndScaleVector(C3DVector(Vertex.Normal.x,Vertex.Normal.y,Vertex.Normal.z)));
                self.pVertexTexCoord.append([Vertex.TextureCoordinate.x,1.0-Vertex.TextureCoordinate.y]);
        # Get the triangle count and the triangles with their material indices
        for f in MeshObject.faces:
            if(VerticesResorted):
                piVertex=ppiTriangleVertex[f.index];
            for i in range(len(f.verts)-1):
                if(i!=len(f.verts)-1-i and i+1!=len(f.verts)-1-i):
                    if(not VerticesResorted):
                        self.pTriangle.append(CTriangleIndices(f.verts[i+1].index,f.verts[i].index,f.verts[len(f.verts)-1-i].index));
                    else:
                        self.pTriangle.append(CTriangleIndices(piVertex[i+1],piVertex[i],piVertex[len(f.verts)-1-i]));
                    if(len(MeshObject.materials)>0):
                        if(MeshObject.materials[f.mat] in MaterialList):
                            self.piTriangleMaterial.append(MaterialList.index(MeshObject.materials[f.mat]));
                        else:
                            self.piTriangleMaterial.append(0);
                    else:
                        self.piTriangleMaterial.append(0);
                    self.nTriangle+=1;
        # Get the skinning weights if necessary
        if(len(BoneList)>0):
            self.SkinningUsed=True;
            self.iMaxBone=len(BoneList)-1;
            for i in range(len(MeshObject.verts)):
                # Get the influences of this vertex
                pInfluence=MeshObject.getVertexInfluences(i);
                # Extend the arrays
                self.ppVertexSkinWeight.append([]);
                self.ppiVertexBone.append([]);
                # Find the four biggest bone influences
                while(len(self.ppVertexSkinWeight[i])<4 and len(pInfluence)>0):
                    MaximumInfluence=0.0;
                    iMaximumInfluence=0;
                    for j in range(len(pInfluence)):
                        if(MaximumInfluence<pInfluence[j][1]):
                            MaximumInfluence=pInfluence[j][1];
                            iMaximumInfluence=j;
                    NewInfluence=pInfluence.pop(iMaximumInfluence);
                    self.ppVertexSkinWeight[i].append(NewInfluence[1]);
                    # Get the bone index
                    BoneFound=False;
                    for j in range(len(BoneList)):
                        if(BoneList[j].Name==NewInfluence[0]):
                            self.ppiVertexBone[i].append(j);
                            BoneFound=True;
                            break;
                    if(BoneFound==False):
                        print("Failed to find a bone called \""+NewInfluence[0]+"\". It is refered to by a vertex in the mesh \""+self.Name+"\". The weight will not be used.");
                        self.ppiVertexBone[i].append(0);
                        self.ppVertexSkinWeight[i].pop();
                        self.ppVertexSkinWeight[i].append(0.0);
                # Make sure that there are at least four skin weights
                while(len(self.ppVertexSkinWeight[i])<4):
                    self.ppVertexSkinWeight[i].append(0.0);
                    self.ppiVertexBone[i].append(0);
                # Make sure that there's not no weight
                TotalWeight=sum(self.ppVertexSkinWeight[i]);
                if(TotalWeight==0.0):
                    TotalWeight=1.0;
                    self.ppVertexSkinWeight[i][0]=1.0;
                # "Normalize" the weights
                for j in range(4):
                    self.ppVertexSkinWeight[i][j]/=TotalWeight;
        # If the vertices have been resorted the skin weights will need to be
        # resorted
        if(VerticesResorted and len(BoneList)>0):
            ppNewVertexSkinWeight=self.nVertex*[[1.0,0.0,0.0,0.0]];
            ppiNewVertexBone=self.nVertex*[[0,0,0,0]];
            for i in range(len(MeshObject.faces)):
                for j in range(len(MeshObject.faces[i].verts)):
                    ppNewVertexSkinWeight[ppiTriangleVertex[i][j]]=self.ppVertexSkinWeight[MeshObject.faces[i].verts[j].index];
                    ppiNewVertexBone[ppiTriangleVertex[i][j]]=self.ppiVertexBone[MeshObject.faces[i].verts[j].index];
            self.ppVertexSkinWeight=ppNewVertexSkinWeight;
            self.ppiVertexBone=ppiNewVertexBone;


    def SaveData(self,File):
        """This function writes the collected data to the given file object"""
        # Check whether the model can be written
        if(self.nVertex>65535 or self.nTriangle>65535):
            print("The model can not be written to an Ultimate 3D model file, because one of the meshes has more than 65535 triangles or vertices.");
            Draw.PupMenu("Error%t|The model can not be exported because one mesh contains more than 65535 triangles/vertices.");
            return False;
        # Write the null-terminated chunk identifier
        File.write("$U3D_MESH");
        File.write(struct.pack("b",0));
        # Write the null-terminated mesh name
        File.write(self.Name);
        File.write(struct.pack("b",0));
        # Write the frame index for the mesh (0)
        File.write(struct.pack("H",0));
        # Write the lod of this mesh (1.0)
        File.write(struct.pack("f",1.0));
        # Write the normal scalar (1.0)
        if(self.SkinningUsed):
            File.write(struct.pack("f",-1.0));
        else:
            File.write(struct.pack("f",1.0));
        # Write whether this mesh contains valid inverse tangent space matrices
        File.write(struct.pack("b",0));
        # Write the vertex count
        File.write(struct.pack("H",self.nVertex));
        # Write the texture coordinate dimension array
        File.write(struct.pack("8H",2*self.bTextureCoordinatesAvailable,0,0,0,0,0,0,0));
        # Write the number of skin weights per vertex
        File.write(struct.pack("H",3*self.SkinningUsed));
        # Write the maximum index of the bones that are influencing this mesh
        File.write(struct.pack("B",self.iMaxBone));
        # Write the vertex and triangle count of shadow optimized geometry
        File.write(struct.pack("HH",0,0));
        # Write the vertex position array
        for Position in self.pVertexPosition:
            Position.SaveData(File);
        # Write the vertex normals
        for Normal in self.pVertexNormal:
            Normal.SaveAsCompressedNormal(File);
        # Whether a texture coordinate buffer is to be used and whether the
        # buffer is owned
        File.write(struct.pack("b",0));
        File.write(struct.pack("b",1));
        # Write the vertex texture coordinates
        if self.bTextureCoordinatesAvailable:
            for TexCoord in self.pVertexTexCoord:
                File.write(struct.pack("f",TexCoord[0]));
                File.write(struct.pack("f",TexCoord[1]));
        # Write the skin weights
        for SkinWeight in self.ppVertexSkinWeight:
            File.write(struct.pack("3f",SkinWeight[0],SkinWeight[1],SkinWeight[2]));
        # Write the bone indices for each vertex
        for BoneIndex in self.ppiVertexBone:
            File.write(struct.pack("4B",BoneIndex[0],BoneIndex[1],BoneIndex[2],BoneIndex[3]));
        # Write the triangle count
        File.write(struct.pack("H",self.nTriangle));
        # Whether the triangle indices are owned by this mesh (true)
        File.write(struct.pack("b",1));
        # Write the index buffer
        for Triangle in self.pTriangle:
            Triangle.SaveData(File);
        # Write the triangle material indices
        for i in range(self.nTriangle):
            File.write(struct.pack("H",self.piTriangleMaterial[i]));
        # The function has succeeded
        return True;


    def AppendMesh(self,MeshObject):
        """This function appends the given mesh object to this mesh object."""
        # Combine the names
        self.Name=self.Name+"+"+MeshObject.Name;
        # Combine the triangle lists
        for i in range(MeshObject.nTriangle):
            self.pTriangle.append(CTriangleIndices(MeshObject.pTriangle[i].A+self.nVertex,MeshObject.pTriangle[i].B+self.nVertex,MeshObject.pTriangle[i].C+self.nVertex));
            self.piTriangleMaterial.append(MeshObject.piTriangleMaterial[i]);
        self.nTriangle+=MeshObject.nTriangle;
        # Combine the vertex lists
        self.pVertexPosition.extend(MeshObject.pVertexPosition);
        self.pVertexNormal.extend(MeshObject.pVertexNormal);
        if(self.bTextureCoordinatesAvailable==False and MeshObject.bTextureCoordinatesAvailable==True):
            self.pVertexTexCoord=[[0.0,0.0]]*self.nVertex;
            self.bTextureCoordinatesAvailable=True;
        if(self.bTextureCoordinatesAvailable):
            if(MeshObject.bTextureCoordinatesAvailable):
                self.pVertexTexCoord.extend(MeshObject.pVertexTexCoord);
            else:
                self.pVertexTexCoord.extend([[0.0,0.0]]*MeshObject.nVertex);
        self.ppVertexSkinWeight.extend(MeshObject.ppVertexSkinWeight);
        self.ppiVertexBone.extend(MeshObject.ppiVertexBone);
        self.nVertex+=MeshObject.nVertex;


    def CleanUp(self):
        """This function cleans up this mesh object completely."""
        self.Name="";
        self.nVertex=self.nTriangle=0;
        self.bTextureCoordinatesAvailable=False;
        self.pVertexPosition=[];
        self.pVertexNormal=[];
        self.pVertexTexCoord=[];
        self.pTriangle=[];
        self.piTriangleMaterial=[];
        self.SkinningUsed=False;
        self.iMaxBone=0;
        self.ppVertexSkinWeight=[];
        self.ppiVertexBone=[];

    # The name of this mesh
    Name="";

    # The number of vertices and triangles
    nVertex=0;
    nTriangle=0;

    # Whether texture coordinates are available
    bTextureCoordinatesAvailable=0;

    # The vertex position array
    pVertexPosition=[];
    # The vertex normal vector array
    pVertexNormal=[];
    # The vertex texture coordinate array
    pVertexTexCoord=[];
    # The number of skin weights per vertex
    SkinningUsed=False;
    iMaxBone=0;
    ppVertexSkinWeight=[];
    ppiVertexBone=[];

    # The triangle index array
    pTriangle=[];
    # The triangle material index array
    piTriangleMaterial=[];


class CMaterial:
    """This class is used to collect data about materials and to write it to
       a *.u3d file."""

    def __init__(self,MaterialObject,TexturePath):
        """This constructor collects data from the given material object."""
        self.CleanUp();
        # Get the material name
        self.Name=MaterialObject.name;
        # Get the material colors
        self.AmbientColor=CColor(MaterialObject.R*MaterialObject.amb,MaterialObject.G*MaterialObject.amb,MaterialObject.B*MaterialObject.amb,MaterialObject.alpha);
        self.DiffuseColor=CColor(MaterialObject.R,MaterialObject.G,MaterialObject.B,MaterialObject.alpha);
        self.SpecularColor=CColor(MaterialObject.specR*MaterialObject.spec,MaterialObject.specG*MaterialObject.spec,MaterialObject.specB*MaterialObject.spec,1.0);
        self.EmissiveColor=CColor(MaterialObject.R*MaterialObject.emit,MaterialObject.G*MaterialObject.emit,MaterialObject.B*MaterialObject.emit,1.0);
        self.SpecularPower=MaterialObject.getHardness();
        # Determine the number of used textures and get the list of texture
        # files (after unpacking them in case they are packed)
        self.nTexture=0;
        for Texture in MaterialObject.getTextures():
            if(Texture==None):
                continue;
            Image=Texture.tex.getImage();
            if(Image==None):
                continue;
            # Unpack the texture if necessary
            Packed=Image.packed;
            if(Image.packed):
                Image.unpack(Blender.UnpackModes["WRITE_LOCAL"]);
            # Get the file name (without pass)
            FileName=Image.getFilename();
            for i in range(len(FileName)-1,0,-1):
                if(FileName[i]=="/" or FileName[i]=="\\"):
                    FileName=FileName[i+1:];
                    break;
            # Remove slashes from the beginning of the file name with path
            FileNameWithPath=Image.getFilename();
            while FileNameWithPath[0]=="\\" or FileNameWithPath[0]=="/":
                FileNameWithPath=FileNameWithPath[1:];
            # Copy the texture file to the texture path
            if(os.path.exists(TexturePath+FileName)):
                try:
                    os.remove(TexturePath+FileName);
                except OSError:
                    print("Failed to remove the recently exported texture file \""+TexturePath+FileName+"\".");
            try:
                shutil.copyfile(FileNameWithPath,TexturePath+FileName);
            except IOError:
                print("Failed to copy the texture file \""+FileNameWithPath+"\" to the gfx folder (\""+TexturePath+FileName+"\"). You may be able to fix this by opening the file directly with Blender instead of opening it through File->Open.");
            # Add the texture to the list
            self.nTexture+=1;
            self.pTextureFile.append(FileName);
            # Repack the texture in case it has been unpacked
            if(Packed):
                Image.pack();
                # Remove the file in case it still exists
                if(os.path.exists(FileNameWithPath)):
                    try:
                        os.remove(FileNameWithPath);
                    except OSError:
                        print("Failed to remove a temporarily unpacked texture file.");


    def SaveData(self,File):
        # Write the null-terminated material identifier
        File.write("$U3D_MAT");
        File.write(struct.pack("b",0));
        # Write the null-terminated material name
        File.write(self.Name);
        File.write(struct.pack("b",0));
        # Write default material colors
        self.AmbientColor.SaveData(File);
        self.DiffuseColor.SaveData(File);
        self.SpecularColor.SaveData(File);
        self.EmissiveColor.SaveData(File);
        # Write the specular power
        File.write(struct.pack("f",self.SpecularPower));
        # Write the number of textures
        self.nTexture=min(8,self.nTexture);
        File.write(struct.pack("H",self.nTexture));
        # Write default texture operations
        if(self.nTexture>0):
            File.write(struct.pack("L",5));
        else:
            File.write(struct.pack("L",1));
        for i in range(1,self.nTexture):
            File.write(struct.pack("L",4));
        for i in range(max(1,self.nTexture),8):
            File.write(struct.pack("L",1));
        # Write default texture coordinate sets
        File.write(struct.pack("8L",0,0,0,0,0,0,0,0));
        # Write whether the resources are part of the file (false)
        File.write(struct.pack("b",0));
        # Write the texture data
        nWrittenTexture=0;
        for TextureFile in self.pTextureFile:
            if(nWrittenTexture==8):
                break;
            File.write(struct.pack("b",True));
            File.write(struct.pack("I",0));
            File.write(TextureFile);
            File.write(struct.pack("b",0));
            File.write(struct.pack("H",0));
            File.write(struct.pack("H",0));
            File.write(TextureFile);
            File.write(struct.pack("b",0));
            nWrittenTexture+=1;
        for i in range(nWrittenTexture,8):
            File.write(struct.pack("b",0));
        # Write that no effect is being used
        File.write(struct.pack("b",0));

    def CleanUp(self):
        """This function cleans up this material object completely."""
        self.Name="";
        self.pTextureFile=[];

    # The name of this material
    Name="";
    # The material colors
    AmbientColor=CColor(1.0,1.0,1.0,1.0);
    DiffuseColor=CColor(1.0,1.0,1.0,1.0);
    SpecularColor=CColor(0.0,0.0,0.0,1.0);
    EmissiveColor=CColor(0.0,0.0,0.0,1.0);
    SpecularPower=0.0;
    # The number of used textures
    nTexture=0;
    # A list of strings containing texture file names
    pTextureFile=[];


class CScalingKey:
    """This class handles a single scaling key."""

    def __init__(self,Frame,Scaling):
        """Initialize this key with the given data."""
        self.Frame=Frame;
        self.Scaling=Scaling;

    def SaveData(self,File):
        """Write the data of this key to the given file."""
        File.write(struct.pack("H",self.Frame));
        File.write(struct.pack("H",0));
        self.Scaling.SaveData(File);

    Frame=0;
    Scaling=C3DVector(0.0,0.0,0.0);


class CTranslationKey:
    """This class handles a single translation key."""

    def __init__(self,Frame,Translation):
        """Initialize this key with the given data."""
        self.Frame=Frame;
        self.Translation=Translation;

    def SaveData(self,File):
        """Write the data of this key to the given file."""
        File.write(struct.pack("H",self.Frame));
        File.write(struct.pack("H",0));
        self.Translation.SaveData(File);
        
    Frame=0;
    Translation=C3DVector(0.0,0.0,0.0);


class CRotationKey:
    """This class handles a single rotation key."""

    def __init__(self,Frame,Rotation):
        """Initialize this key with the given data."""
        self.Frame=Frame;
        self.Rotation=Rotation;

    def SaveData(self,File):
        """Write the data of this key to the given file."""
        File.write(struct.pack("H",self.Frame));
        File.write(struct.pack("H",0));
        self.Rotation.SaveData(File);

    Frame=0;
    Rotation=CQuaternion(0.0,0.0,0.0,1.0);


class CBone:
    """This class is used to collect data about a single bone and to write it
       this data to a *.u3d file."""

    def Initialize(self,BoneObject,BlenderBoneList,pAction,ExportThroughIpo,ArmatureObj):
        """This constructor collects data from the given bone. Indices will be
           taken from the given bone list (offset with one)."""
        self.CleanUp();
        # Get the name
        self.Name=BoneObject.name;
        # Find the parent bone (and in particular get it's index)
        if(BoneObject.hasParent()):
            self.iParent=BlenderBoneList.index(BoneObject.parent)+1;
        else:
            self.iParent=0;
        # Get the mesh transformation
        self.MeshTransformation.CreateFromBlenderMatrix(ArmatureObj.getMatrix().copy().invert()*BoneObject.matrix["ARMATURESPACE"].copy().invert());
        # Determine the bone to parent bone space matrix
        if(self.iParent>0):
            BoneToParentSpace=BoneObject.matrix["ARMATURESPACE"]*(BlenderBoneList[self.iParent-1].matrix["ARMATURESPACE"].copy().invert());
        else:
            BoneToParentSpace=BoneObject.matrix["ARMATURESPACE"]*ArmatureObj.getMatrix();
        # Create the keys
        for Action in pAction:
            if(len(Action.getFrameNumbers())==0):
                continue;
            if(ExportThroughIpo):
                self.AddKeysFromActionIpo(Action,BoneToParentSpace);
            else:
                self.AddKeysFromActionPose(Action,ArmatureObj);
            self.nFrame+=max(Action.getFrameNumbers())+1;
        # Make sure that there's at least one key of every type to have the
        # bone to parent space transformation included
        if(len(self.pTranslationKey)==0):
            BoneToParentTranslation=BoneToParentSpace.translationPart();
            self.pTranslationKey.append(CTranslationKey(0,C3DVector(BoneToParentTranslation[0],BoneToParentTranslation[1],BoneToParentTranslation[2])));
        if(len(self.pRotationKey)==0):
            BoneToParentRotation=BoneToParentSpace.toQuat();
            self.pRotationKey.append(CRotationKey(0,CQuaternion(BoneToParentRotation.x,BoneToParentRotation.y,BoneToParentRotation.z,BoneToParentRotation.w)));
        if(len(self.pScalingKey)==0):
            BoneToParentScaling=BoneToParentSpace.scalePart();
            self.pScalingKey.append(CScalingKey(0,C3DVector(BoneToParentScaling.x,BoneToParentScaling.y,BoneToParentScaling.z)));


    def AddKeysFromActionIpo(self,Action,BoneToParentSpace):
        """This function adds key frames from the given action to this bone.
           oFrame will be added to the frame values of the key frames."""
        # Get the relevant Ipo
        try:
            AnimationIpo=Action.getChannelIpo(self.Name);
        except:
            # There's no animation for this bone, return
            return;
        # Get the transformation curves in the ipo
        TranslationXCurve=TranslationYCurve=TranslationZCurve=None;
        RotationXCurve=RotationYCurve=RotationZCurve=RotationWCurve=None;
        ScalingXCurve=ScalingYCurve=ScalingZCurve=None;
        for Curve in AnimationIpo:
            if(Curve.name=="PO_LOCX" or Curve.name=="LocX"):
                TranslationXCurve=Curve;
            elif(Curve.name=="PO_LOCY" or Curve.name=="LocY"):
                TranslationYCurve=Curve;
            elif(Curve.name=="PO_LOCZ" or Curve.name=="LocZ"):
                TranslationZCurve=Curve;
            elif(Curve.name=="PO_QUATX" or Curve.name=="QuatX"):
                RotationXCurve=Curve;
            elif(Curve.name=="PO_QUATY" or Curve.name=="QuatY"):
                RotationYCurve=Curve;
            elif(Curve.name=="PO_QUATZ" or Curve.name=="QuatZ"):
                RotationZCurve=Curve;
            elif(Curve.name=="PO_QUATW" or Curve.name=="QuatW"):
                RotationWCurve=Curve;
            elif(Curve.name=="PO_SIZEX" or Curve.name=="SizeX"):
                ScalingXCurve=Curve;
            elif(Curve.name=="PO_SIZEY" or Curve.name=="SizeY"):
                ScalingYCurve=Curve;
            elif(Curve.name=="PO_SIZEZ" or Curve.name=="SizeZ"):
                ScalingZCurve=Curve;
        # Create the keys (the transformation described by them needs to be
        # "multiplied from the right" with the bone to parent space matrix)
        SortedKeyFrameList=Action.getFrameNumbers();
        SortedKeyFrameList.sort();
        BoneToParentScaling=BoneToParentSpace.scalePart();
        # If the scaling is not homogenous output a warning
        if(BoneToParentScaling.x<BoneToParentScaling.y-0.001 or BoneToParentScaling.x>BoneToParentScaling.y+0.001 or BoneToParentScaling.y<BoneToParentScaling.z-0.001 or BoneToParentScaling.y>BoneToParentScaling.z+0.001):
            print("WARNING: The bone "+self.Name+" has an inhomogenous scaling in relation to its parent. Since inhomogenous scaling (scaling that's not the same for all axes) is not supported the scaling has been made homogenous.");
            print("The inhomogenous scaling is X="+str(BoneToParentScaling.x)+" Y="+str(BoneToParentScaling.y)+" Z="+str(BoneToParentScaling.z));
            BoneToParentScaling.x=BoneToParentScaling.y=BoneToParentScaling.z=(BoneToParentScaling.x+BoneToParentScaling.y+BoneToParentScaling.z)/3.0;
        BoneToParentTranslation=BoneToParentSpace.translationPart();
        BoneToParentRotation=BoneToParentSpace.toQuat();
        BoneToParentRotation=CQuaternion(BoneToParentRotation.x,BoneToParentRotation.y,BoneToParentRotation.z,BoneToParentRotation.w);
        for i in SortedKeyFrameList:
            KeyTranslation=C3DVector(0.0,0.0,0.0);
            KeyScaling=C3DVector(1.0,1.0,1.0);
            KeyRotation=CQuaternion(0.0,0.0,0.0,1.0);
            if(TranslationXCurve!=None):
                KeyTranslation=C3DVector(TranslationXCurve[i],TranslationYCurve[i],TranslationZCurve[i]);
            if(RotationXCurve!=None):
                KeyRotation=CQuaternion(RotationXCurve[i],RotationYCurve[i],RotationZCurve[i],RotationWCurve[i]);
            if(ScalingXCurve!=None):
                KeyScaling=C3DVector(ScalingXCurve[i],ScalingYCurve[i],ScalingZCurve[i]);
            # Transform the key data
            KeyScaling=C3DVector(KeyScaling.x*BoneToParentScaling.x,KeyScaling.y*BoneToParentScaling.y,KeyScaling.z*BoneToParentScaling.z);
            RotatedTranslation=BoneToParentRotation.RotateVector(KeyTranslation.x,KeyTranslation.y,KeyTranslation.z);
            KeyTranslation=C3DVector(RotatedTranslation.x+BoneToParentTranslation.x,RotatedTranslation.y+BoneToParentTranslation.y,RotatedTranslation.z+BoneToParentTranslation.z);
            KeyRotation.AppendRotation(BoneToParentRotation);
            # Save the transformed key data
            if(TranslationXCurve!=None):
                self.pTranslationKey.append(CTranslationKey(i+self.nFrame,KeyTranslation));
            if(RotationXCurve!=None):
                self.pRotationKey.append(CRotationKey(i+self.nFrame,KeyRotation));
            if(ScalingXCurve!=None):
                self.pScalingKey.append(CScalingKey(i+self.nFrame,KeyScaling));


    def AddKeysFromActionPose(self,Action,ArmatureObj):
        """This function implements an important feature in the only possible
           way, which is very ugly. It makes it possible to retrieve
           animations, which use constraints like IKs. The keys will be appended
           to the array of this instance."""
        # Get the pose
        Pose=ArmatureObj.getPose();
        # Save the current pose and clear the pose bone transformations
        pPoseBoneTranslation=[];
        pPoseBoneScaling=[];
        pPoseBoneRotation=[];
        for PoseBone in Pose.bones.values():
            pPoseBoneTranslation.append(PoseBone.loc);
            pPoseBoneScaling.append(PoseBone.size);
            pPoseBoneRotation.append(PoseBone.quat);
            PoseBone.loc.x=PoseBone.loc.y=PoseBone.loc.z=0.0;
            PoseBone.size.x=PoseBone.size.y=PoseBone.size.z=1.0;
            PoseBone.quat.x=PoseBone.quat.y=PoseBone.quat.z=0.0;
            PoseBone.quat.w=1.0;
        # Save some information that needs to be restored
        PreviousAction=ArmatureObj.getAction();
        PreviousFrame=Blender.Get("curframe");
        # Activate the given action
        Action.setActive(ArmatureObj);
        # Cycle through all relevant frames
        # NOTE: It may be that additional frames are relevant. An alternative,
        # but also bad solution would be exporting one key for every frame.
        SortedKeyFrameList=Action.getFrameNumbers();
        if(0 not in SortedKeyFrameList):
            SortedKeyFrameList.append(0);
        SortedKeyFrameList.sort();
        for Frame in SortedKeyFrameList:
            # Set up the right frame for the pose
            ArmatureObj.evaluatePose(Frame);
            # Retrieve the transformation in relation to the parent
            PoseBoneTransformation=Pose.bones[self.Name].poseMatrix;
            if(Pose.bones[self.Name].parent!=None):
                PoseBoneTransformation*=Pose.bones[self.Name].parent.poseMatrix.copy().invert();
            else:
                PoseBoneTransformation*=ArmatureObj.getMatrix();
            # Create the transformation keys
            Rotation=PoseBoneTransformation.toQuat();
            Scaling=PoseBoneTransformation.scalePart();
            Translation=PoseBoneTransformation.translationPart();
            RotationFactor=1;
            if(len(self.pRotationKey)>0):
                PreviousRotation=self.pRotationKey[len(self.pRotationKey)-1].Rotation;
                if(PreviousRotation.x*Rotation.x+PreviousRotation.y*Rotation.y+PreviousRotation.z*Rotation.z+PreviousRotation.w*Rotation.w<0.0):
                    RotationFactor=-1;
            self.pRotationKey.append(CRotationKey(Frame+self.nFrame,CQuaternion(Rotation.x*RotationFactor,Rotation.y*RotationFactor,Rotation.z*RotationFactor,Rotation.w*RotationFactor)));
            self.pScalingKey.append(CScalingKey(Frame+self.nFrame,C3DVector(Scaling.x,Scaling.y,Scaling.z)));
            self.pTranslationKey.append(CTranslationKey(Frame+self.nFrame,C3DVector(Translation.x,Translation.y,Translation.z)));
        # Restore the pose transformation
        for PoseBone in Pose.bones.values():
            PoseBone.loc=pPoseBoneTranslation.pop(0);
            PoseBone.size=pPoseBoneScaling.pop(0);
            PoseBone.quat=pPoseBoneRotation.pop(0);
        # Restore the previous action and the previous frame
        if(PreviousAction!=None):
            PreviousAction.setActive(ArmatureObj);
        else:
            ArmatureObj.action=None;
        Blender.Set("curframe",PreviousFrame);


    def SaveData(self,File):
        """This function saves all data of this bone to the given file."""
        # Write the null-terminated bone chunk identifier
        File.write("$U3D_BONE");
        File.write(struct.pack("B",0));
        # Write the null-terminated name of this bone
        File.write(self.Name);
        File.write(struct.pack("B",0));
        # Write the index of the parent of this bone
        File.write(struct.pack("H",self.iParent));
        # Write the bone frame offset (-1.0) and whether this is to effect the
        # child frame
        File.write(struct.pack("f",-1.0));
        File.write(struct.pack("B",0));
        # Write the number of meshes that are appended to this bone
        File.write(struct.pack("H",1));
        # Write the index of the skinning mesh
        File.write(struct.pack("H",0));
        # Write the mesh transformation
        self.MeshTransformation.SaveData(File);
        # Write the scaling keys
        File.write(struct.pack("H",len(self.pScalingKey)));
        for ScalingKey in self.pScalingKey:
            ScalingKey.SaveData(File);
        # Write the translation keys
        File.write(struct.pack("H",len(self.pTranslationKey)));
        for TranslationKey in self.pTranslationKey:
            TranslationKey.SaveData(File);
        # Write the rotation keys
        File.write(struct.pack("H",len(self.pRotationKey)));
        for RotationKey in self.pRotationKey:
            RotationKey.SaveData(File);


    def MakeBlenderToU3DSpaceBone(self):
        self.CleanUp();
        # Set up a descriptive name...
        self.Name="BlenderToU3DSpaceBone";
        # ...no parent...
        self.iParent=65535;
        # ...an identity mesh transformation...
        self.MeshTransformation=CMatrix4x4(1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0);
        # ...and keys that describe the BlenderToU3DSpace matrix.
        self.pScalingKey=[CScalingKey(0,C3DVector(1.0,-1.0,1.0))];
        self.pRotationKey=[CRotationKey(0,CQuaternion(1.0,0.0,0.0,-1.0))];
        self.pTranslationKey=[];
        self.nFrame=0;


    def CleanUp(self):
        """This function erases all data that's saved within this bone."""
        self.Name="";
        self.iParent=0;
        self.MeshTransformation=CMatrix4x4(1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0);
        self.pScalingKey=[];
        self.pRotationKey=[];
        self.pTranslationKey=[];
        self.nFrame=0;
        

    # The name of this bone
    Name="";
    # The index of the parent bone
    iParent=0;
    # The transformation that is to be applied to the geometry before the 
    # final bone transformation
    MeshTransformation=CMatrix4x4(1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0);
    # The scaling keys of this bone
    pScalingKey=[];
    # The translation keys of this bone
    pRotationKey=[];
    # The rotation keys of this bone
    pTranslationKey=[];
    # The length of the animation of this bone
    nFrame=0;


class CSkeleton:
    """This class is used to collect data about all armatures in a given scene
       and to write it to a *.u3d file."""

    def __init__(self,RelevantObjects,ExportConstraints,ExportActionList,ActionListFileName):
        """This init function initializes this skeleton from the armatures in
           the given scene."""
        # Create the Blender to U3D space root bone
        self.BoneList=[CBone()];
        self.BoneList[0].MakeBlenderToU3DSpaceBone();
        # Create a list of absolutely all bones in the scene
        for Obj in RelevantObjects:
            if(Obj.getType()=="Armature"):
                # Create a list of all blender bones
                iFirstArmatureBone=len(self.BlenderBoneList);
                for Bone in Obj.getData().bones.values():
                    self.BlenderBoneList.append(Bone);
                # Collect the needed information from the bones
                for i in range(iFirstArmatureBone,len(self.BlenderBoneList)):
                    self.BoneList.append(CBone());
                    self.BoneList[i+1].Initialize(self.BlenderBoneList[i],self.BlenderBoneList,Armature.NLA.GetActions().values(),not ExportConstraints,Obj);
        # Output the association between frames and actions and determine the
        # number of frames that is needed
        self.nFrame=0;
        if(ExportActionList):
            ActionListFile=open(ActionListFileName,"w");
        for Act in Armature.NLA.GetActions().values():
            if(len(Act.getFrameNumbers())==0):
                continue;
            nActionFrame=max(Act.getFrameNumbers())+1;
            if(nActionFrame>1):
                print("Action \""+Act.name+"\" takes the range from frame "+str(self.nFrame)+" to frame "+str(self.nFrame+nActionFrame-1)+".");
                if(ExportActionList):
                    ActionListFile.write("\""+Act.name+"\" "+str(self.nFrame)+" "+str(self.nFrame+nActionFrame-1)+"\n");
            self.nFrame+=nActionFrame;
        self.nFrame=max(1,self.nFrame);
        if(ExportActionList):
            ActionListFile.close();


    def SaveData(self,File):
        """Saves the skeleton to the given file."""
        for Bone in self.BoneList:
            Bone.SaveData(File);
        

    # A list of all Blender bone objects in the scene
    BlenderBoneList=[];
    # A list of Ultimate 3D bone objects
    BoneList=[];
    # The number of frames in the animation of this skeleton
    nFrame=1;


def FileSelectorCallback(FileName):
    """This is the callback function for the file selector"""
    ExportPlugin.InputCompleted(FileName);


class CInputWindow:
    """This class creates a window in which the user has to enter
       information about how he wants the file to be exported."""

    def __init__(self):
        """This function creates a user interface to get the input."""
        # Display a neat little pop up with the options
        FormContent=[];
        FormContent.append(("Selection only",self.ExportSelectionOnly,"If checked only selected objects will be exported, otherwise the whole scene will be exported."));
        FormContent.append(("Animation",self.ExportArmatures,"Export armatures, skin weights and acions?"));
        FormContent.append(("Constraints",self.ExportConstraints,"Cycle through all frames to export constraints?"));
        FormContent.append(("Remove double vertices",self.RemoveDoubleVertices,"Remove double vertices? This can take a while for models with many vertices."));
        FormContent.append(("Action list",self.ExportActionList,"Export a list of all actions with their corresponding frame values?"));
        PopUpReturnValue=Draw.PupBlock("Ultimate 3D export settings",FormContent);
        # If OK has been clicked show the file selection dialog
        if(PopUpReturnValue):
            # Suggest a filename that makes sense
            ExportFileSuggestion=Blender.Get("filename");
            print(ExportFileSuggestion);
            for i in range(len(ExportFileSuggestion)-1,0,-1):
                if(ExportFileSuggestion[i]=="."):
                    ExportFileSuggestion=ExportFileSuggestion[0:i]+".u3d";
                    break;
            if(len(ExportFileSuggestion)==0):
                ExportFileSuggestion="Model.u3d";
            Window.FileSelector(FileSelectorCallback,"Export *.u3d model",ExportFileSuggestion);

    ExportSelectionOnly=Draw.Create(False);
    ExportArmatures=Draw.Create(True);
    ExportConstraints=Draw.Create(True);
    RemoveDoubleVertices=Draw.Create(False);
    ExportActionList=Draw.Create(False);


class CExportPlugin:
    """This class runs the whole export plugin."""

    def __init__(self):
        """As the constructor gets called the plugin starts working."""
        self.InputWindow=CInputWindow();

    def InputCompleted(self,FileName):
        """When the input window has retrieved all input this function gets
           called."""
        if(len(FileName)>3):
            self.ExportModel(FileName,self.InputWindow.ExportSelectionOnly.val,self.InputWindow.ExportArmatures.val,self.InputWindow.ExportConstraints.val,self.InputWindow.RemoveDoubleVertices.val,self.InputWindow.ExportActionList.val);

    def ExportModel(self,FileName,ExportSelectionOnly,ExportArmatures,ExportConstraints,RemoveDoubleVertices,ExportActionList):
        """This function initializes all processes that are necessary to export
           the model to a given file."""
        print("*** Exporting to an Ultimate 3D model file.***");

        # Prepare some information here:
        # Get the relevant scene
        SceneObject=Scene.GetCurrent();
        # Create a list of all objects (including those, which are in groups)
        AllObjects=[];
        for Obj in SceneObject.objects:
            AllObjects.append(Obj);
        for GroupObj in Group.Get():
            for Obj in GroupObj.objects:
                AllObjects.append(Obj);
        # Create a list of the objects, which may need to be exported (if their
        # type is right)
        RelevantObjects=[];
        if(ExportSelectionOnly):
            for Obj in AllObjects:
                if(Obj.isSelected()):
                    RelevantObjects.append(Obj);
            if(len(RelevantObjects)==0):
                print("Error: No object selected for exporting.");
                Draw.PupMenu("Error%t|No objects selected for exporting.");
                return False;
        else:
            for Obj in AllObjects:
                RelevantObjects.append(Obj);
            if(len(RelevantObjects)==0):
                print("Error: No object contained in the scene.");
                Draw.PupMenu("Error%t|No object contained in the scene.");
 
        # Sort out meshes, which do not contain any geometry and count
        # the meshes
        nMesh=0;
        for Object in RelevantObjects:
            if(Object.getType()=="Mesh"):
                # Count the number of triangles
                nTriangle=0;
                for Face in Object.getData(False,True).faces:
                    nTriangle+=len(Face.verts);
                # Remove the mesh if there are zero triangles
                if nTriangle==0:
                    RelevantObjects.remove(Object);
                else:
                    nMesh+=1;
        # The model is invalid, if it does not contain any meshes
        if(nMesh==0):
            print("The model you are trying to export does not contain any meshes. This is invalid. Every model has to contain at least one mesh with at least one triangle. Please add some dummy mesh and try again.");
            Draw.PupMenu("Error%t|No geometry found in the model. Please add at least one triangle.");
            return False;

        # Create a list of all materials (without double materials)
        MaterialList=[];
        for Object in RelevantObjects:
            if(Object.getType()=="Mesh"):
                for Material in Object.getData().materials:
                    if(Material not in MaterialList):
                        MaterialList.append(Material);

        # Check whether there's an armature in the scene. If so skinning will
        # be used
        SkinningUsed=False;
        for Object in RelevantObjects:
            if(Object.getType()=="Armature"):
                SkinningUsed=ExportArmatures;
                break;

        # Create a skeleton object if necessary and get it's bone list
        BoneList=[];
        if(SkinningUsed):
            print("Vertex skinning is being used.");
            ActionListFileName="";
            if(ExportActionList):
                ActionListFileName=FileName[0:-3]+"txt";
            Skeleton=CSkeleton(RelevantObjects,ExportConstraints,ExportActionList,ActionListFileName);
            BoneList=Skeleton.BoneList;

        if(len(BoneList)>=128):
            print("Your scene contains more than 126 bones. Since Ultimate 3D and the Ultimate 3D model file format can't handle that many bones in combination with skinning you have to reduce the number of bones. The exporter will terminate now. No file has been created.");
            Draw.PupMenu("Error%t|Too many bones found. There mustn't be more than 126.");
            return False;

        # Try to open the file
        File=open(FileName,"wb");
        if File.closed:
            print("Failed to open the file \""+FileName+"\". Please make sure that you have write access and try again.");
            Draw.PupMenu("Error%t|Failed to open the file \""+FileName+"\".");
            return False;
        print("Opened the file "+FileName+" successfully.");

        # Export the header
        if SkinningUsed:
            HeaderOut=CHeader(RelevantObjects,MaterialList,len(BoneList),Skeleton.nFrame);
        else:
            HeaderOut=CHeader(RelevantObjects,MaterialList,0,1);
        HeaderOut.SaveData(File);

        # Export the meshes
        if(not SkinningUsed):
            # If no skinning is used export each mesh as one mesh chunk
            for i in RelevantObjects:
                if(i.getType()=="Mesh"):
                    ObjectMatrix=i.getMatrix();
                    MeshTransformation=CMatrix4x4(ObjectMatrix[0][0],ObjectMatrix[0][1],ObjectMatrix[0][2],ObjectMatrix[0][3],ObjectMatrix[1][0],ObjectMatrix[1][1],ObjectMatrix[1][2],ObjectMatrix[1][3],ObjectMatrix[2][0],ObjectMatrix[2][1],ObjectMatrix[2][2],ObjectMatrix[2][3],ObjectMatrix[3][0],ObjectMatrix[3][1],ObjectMatrix[3][2],ObjectMatrix[3][3]);
                    MeshOut=CMesh(i.getData(False,True),MaterialList,BoneList,MeshTransformation,RemoveDoubleVertices);
                    if(not MeshOut.SaveData(File)):
                        File.close();
                        return;
            print(str(HeaderOut.nMesh)+" meshes have been exported.");
        else:
            # If skinning is used combine all meshes into one mesh
            MeshList=[];
            for i in RelevantObjects:
                if(i.getType()=="Mesh"):
                    ObjectMatrix=i.getMatrix();
                    MeshTransformation=CMatrix4x4(ObjectMatrix[0][0],ObjectMatrix[0][1],ObjectMatrix[0][2],ObjectMatrix[0][3],ObjectMatrix[1][0],ObjectMatrix[1][1],ObjectMatrix[1][2],ObjectMatrix[1][3],ObjectMatrix[2][0],ObjectMatrix[2][1],ObjectMatrix[2][2],ObjectMatrix[2][3],ObjectMatrix[3][0],ObjectMatrix[3][1],ObjectMatrix[3][2],ObjectMatrix[3][3]);
                    MeshList.append(CMesh(i.getData(False,True),MaterialList,BoneList,MeshTransformation,RemoveDoubleVertices));
            for i in range(1,len(MeshList)):
                MeshList[0].AppendMesh(MeshList[i]);
            if(not MeshList[0].SaveData(File)):
                File.close();
                return;
            print("All meshes in the scene have been combined and the single resulting mesh has been exported.");

        # Export the material list
        for BlenderMaterial in MaterialList:
            TexturePath=FileName;
            for i in range(len(FileName)-1,0,-1):
                if(TexturePath[i]=="/" or TexturePath[i]=="\\"):
                    TexturePath=TexturePath[0:i]+"/gfx/";
                    break;
            if(not os.path.exists(TexturePath)):
                os.mkdir(TexturePath);
            MaterialOut=CMaterial(BlenderMaterial,TexturePath);
            MaterialOut.SaveData(File);
        if(len(MaterialList)):
            print(str(HeaderOut.nMaterial)+" materials have been exported.");
        # If there's no material in the material list create a default material
        else:
            # Material chunk identifier (null-terminated)
            File.write("$U3D_MAT");
            File.write(struct.pack("b",0));
            # Material name (null-terminated)
            File.write("DefaultMaterial");
            File.write(struct.pack("b",0));
            # Ambient color
            File.write(struct.pack("4f",1.0,1.0,1.0,1.0));
            # Diffuse color
            File.write(struct.pack("4f",1.0,1.0,1.0,1.0));
            # Specular color
            File.write(struct.pack("4f",0.0,0.0,0.0,1.0));
            # Emissive color
            File.write(struct.pack("4f",0.0,0.0,0.0,1.0));
            # Specular power
            File.write(struct.pack("f",0.0));
            # Write the number of textures
            File.write(struct.pack("H",0));
            # Write the texture operations
            File.write(struct.pack("8I",5,5,5,5,5,5,5,5));
            # Write the texture coordinate indices
            File.write(struct.pack("8I",0,0,0,0,0,0,0,0));
            # Write that resources don't get exported
            File.write(struct.pack("b",0));
            # Write that there's no texture
            File.write(struct.pack("8b",0,0,0,0,0,0,0,0));
            # Write that there's no effect
            File.write(struct.pack("b",0));
            print("The blender model didn't contain any materials. A default material has been exported.");

        # Export the skeleton
        if(SkinningUsed):
            Skeleton.SaveData(File);
            print("The skeleton has been exported successfully.");
        
        # Close the file
        File.close();

        print("All data has been exported successfully.");
        Draw.PupMenu("Success%t|The model has been exported successfully.");
        
        return True;


# Initialize an instance of CExportPlugin to run the plugin
ExportPlugin=CExportPlugin();
